%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                                                    %
%  This is program for simulating the influence of external field on charge carriers in conduction   %
%  body. It was writen in freeware programming language called Octave, which has basically the same  %
%  synthax as Matlab. The only difference I'm aware of is that Matlab doesn't support definition of  %
%  functions in one script with the body of the program. They must be declared in separate files.    %
%                                                                                                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%used functions in alphabethical order

function ret=bound(a,vec,dvec)  %Limits the movement of charge near the body surface boundary.
  ret=vec+dvec;          %intendet movement (it's out of body)
  scale=norm(dvec);      %scale of intendet step 
  
  ipoint=intersect(vec,dvec);   %finds intersection point with the surface of the body
  nvec=surfn(a,ipoint);         %computes unit normal vector in point of intersection
  
  ret=intersect(ret-scale*nvec, scale*nvec) - 0.1*a*nvec ;  %makes the move back close under the surface  
end


function ret=chvel(a, vel, evec, N)  %Computes the changed velocity of negative charge according to electric field, 
                                     %inertia, disipative forces and boud of step                                     
  lambda=0.01;    %coefficient for step bound stiffnes                                   
  for n=1:N
    dv=0.5*a*evec(:,n)/norm(evec(:,n))*(1- exp( -lambda*norm(evec(:,n)) ) );  %change of velocity according to
                        %electric field direction, its magnituge and step boud stiffnes
    ret(:,n)=0.5*vel(:,n) + dv ;  %accounting inertia and energy dissipation
  end 
end


function ret=closeint(R, r) %Close interaction electric field. 
  ret=r/(R)^3;              %Acts like homogenous ball of charge with radius R.
end


function ret=coulomb(vec)   %Simple coulomb force.
  ret=vec/norm(vec)^3;
end


function ret=cycle(arg)  %Auxiliray function for cycling the agrument of function for view change for plots.
  while arg > 360
    arg = arg -360;
  end
  while arg < 0
    arg = arg + 360;
  end
  ret=arg;
end


function ret=E(vec,neg, pos)   %Computes the electric field in point vec.
  global extfield
  ret=zeros(3,1);
  N=size(pos,2);
  for n=1:N
    r=pos(:,n)-vec;
    ret=ret-coulomb(r);
    r=neg(:,n)-vec;
    ret=ret+coulomb(r);
  end
  ret=ret+extfield;
end


function evec=efield(a,neg,pos, N)  %Computes electric field for every negative charge
  global extfield 
  
  for n(1)=1:N
    evec(:,n(1))=zeros(3,1) - extfield;   %adds the external field
    for n(2)=1:N
      r=pos(:,n(2))-neg(:,n(1));          %relative position of positive charge accorging to given negative charge
      if norm(r) > a
        evec(:,n(1))=evec(:,n(1))+coulomb(r);  %if the pos charge is far enough it interacts with Coulomb force
      else
        evec(:,n(1))=evec(:,n(1))+closeint(a,r); %term for close interaction
      end
      
      r=neg(:,n(2))-neg(:,n(1));
      if n(1) != n(2)
        evec(:,n(1))=evec(:,n(1))-coulomb(r); %coulomb force from other negative charges
      end
    end
  end
end


function lines=forcelin(neg,pos,points)  %Computes the lines of force.
  dr=0.05;
  p=30;

  s=size(points,2);         %number of starting points
  for n(1)=1:s
    point=points(:,n(1));
    lines(:,1,n(1))=point;
    for n(2)=2:p
      Evec= - E(point,neg, pos);      %computes the vector of electric field in given point
      point=point+dr*Evec/norm(Evec); %makes the move in the direction of electric field
      lines(:,n(2),n(1))=point;       
    end
  end
end


function ret=inbody(vec) %Checks if the point vec is inside the body
  ret=0;
  if quad(vec) < 1
    ret=1;
  end  
end


function ret=intersect(vec,dvec) %Computes intersection of the vector dvec from inner point vec with body surface.
  m=0.5;
  for n=2:8
    ret=vec+m*dvec;
    if inbody(ret)
      m=m+0.5^n;
    else
      m=m-0.5^n;
    end       
  end
  ret=vec+m*dvec;
end


function ret=move(a,neg,vel, N)  %Makes the move in unit time step according to velocity of charges
  for n=1:N
    auxneg=neg(:,n)+vel(:,n);
    if inbody(auxneg)
      ret(:,n)=auxneg;
    else  
      ret(:,n)=bound(a,neg(:,n),vel(:,n));   %implements the surface boundary of the body      
    end  
  end
end


function ret=norm(vec)   %Computes norm of the vector.
  ret=sqrt(vec'*vec);    
end


function out(neg,pos,ord)   %Creates the output
  axlim=[-2,2,-1.5,1.5,-1.5,1.5];    %axis limits  
  angles=[cycle(2*ord),10];  %first is rotation around z-axis, second is angle from xy-plane
 
  plot3(neg(1,:),neg(2,:),neg(3,:),'o')  %plots the negative charges
   
    hold on
  plot3(pos(1,:),pos(2,:),pos(3,:),"@21") %plots the positive charges
  
  points=spoints(angles);     %computes starting points for lines of force
  
  lines=forcelin(neg,pos,points);  %computes lines of force
  N=size(lines,3);
  for n=1:N
    plot3(lines(1,:,n),lines(2,:,n),lines(3,:,n),"g")   %plots lines of force
  end

  axis(axlim)                %sets limist for axes
  view(angles(1),angles(2))  %number one view 90 minus arg1, number two view is arg2 
  xlabel("x");
  ylabel("y");
  zlabel("z");
  drawnow;
  print(['plots/plot',int2str(ord),'.png']);  %creates file output; folder plots must be manually created
    
    hold off
end


function ret=points(N,space)         %Generates points from inside of the body in given bounds
  dr=(space(:,2)-space(:,1))./(N-1); %and with given density of the points.
  i=1;
  for n(1)=0:N(1)-1
    for n(2)=0:N(2)-1
      for n(3)=0:N(3)-1
        point=space(:,1)+n'.*dr;
        if inbody(point) 
          ret(:,i)=point;
          i=i+1;
        end
      end
    end
  end
end


function ret=quad(vec)   %Computes the value (x^2/a^2 + y^2/b^2 + z^2/c^2) where [a,b,c] are halfaxes of the quadric
  halfax=[1;0.5;0.5];        %surface.
  ret=vec./halfax;
  ret=ret'*ret;  
end


function ret=spoints(angles)    %generates starting points for drawing the lines of force
  R=1.2;
  N=20;
  phi=angles(1)*(pi/180);
  theta=[1:N]/N*2*pi;

  ret(1,:)=R*cos(phi)*cos(theta);   %The starting points are mounted on circele which axis is oriented 
  ret(2,:)=R*sin(phi)*cos(theta);   %in the view direction.
  ret(3,:)=R*sin(theta);
end


function ret=surfn(a,ipoint)         %Computes surface normal unit vector from point of surface ipoint, 
                                     %oriented out of the body. 
  poi=ipoint*ones(1,3)+0.1*a*eye(3); %generates near points in x,y and z directions                         
  for n=1:3
    ret(n,1)=quad(poi(:,n))-1;       %computes gradient of the quadric surface 
  end
  ret=ret/norm(ret);                 
  if inbody(ipoint+a*ret)
    ret=-ret;                        %correction of orientation
  end  
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%program body

global extfield
extfield=[10;0;10]/sqrt(2);             %homogenous external field

N=8*[1;1;1];                            %number of points in x,y, and z direction 
space=[-1,1;-1,1;-1,1];                 %boundary of the area from which the points of the body are generated
a=min((space(:,2)-space(:,1))./(N-1));  %the minimal distance between positive cores

pos=points(N,space); %generates some points from the inside of the body
Npoi=size(pos,2);    %number of charges (negative and positive sepparately)
neg=pos;             %initially the negative charges are placed at the positive core position
vel=0*neg;           %initial velocity for negative charges

K=300;     %number of snapshots
for k=1:K

  evec=efield(a,neg,pos, Npoi);  %computes resulting electric intensity for every negative charge
  vel=chvel(a, vel, evec, Npoi); %computes velocity resulting from electric field and inertia
  neg=move(a,neg,vel, Npoi);     %moves the negative charges taking bound of the body surface into account

  out(neg,pos,k);                %creates the output
end

